/* 
 *  simple audio Lib .. 
 *  copyrights (c) 1994,1995,1996 by Michael Hipp
 *  SGI audio support copyrights (c) 1995 by Thomas Woerner
 *
 *  A separate Win32 version by Petteri Kangaslampi, pekangas@sci.fi
 *
 *  copy policy: GPL V2.0
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mpg123.h"


/* Buffer length: (global, -1 = default) */
int win32AudioBufferLength = -1;

/* Default buffer length in milliseconds: */
#define BUFFERLEN 500

/* Buffer block size in bytes: */
#define BLOCKSIZE 4096

typedef unsigned char uchar;

static HWAVEOUT waveHandle;
/*static uchar    *blocks[MAXBUFBLOCKS];
static WAVEHDR  *blockHeaders[MAXBUFBLOCKS];
static int      blockPrepared[MAXBUFBLOCKS];*/
static uchar    **blocks;
static WAVEHDR   **blockHeaders;
static int      *blockPrepared;
static unsigned numBlocks;


static int      audioInitialized = 0;

static unsigned blockLen;
static unsigned blockNum, blockPos;

static unsigned sampleRate = 44100;
static unsigned stereo = 1;
static unsigned bits = 16;


static void WinError(MMRESULT error)
{
    fprintf(stderr, "Win32 Wave Output: Error %u\n", (unsigned) error);
    exit(EXIT_FAILURE);
}



static void StopAudio(void)
{
    MMRESULT mmError;
    unsigned i;
    int allDone;
    
    /* Reset wave output device, stop playback, and mark all blocks done: */
    if ( (mmError = waveOutReset(waveHandle)) != 0 )
        WinError(mmError);

    /* Make sure all blocks are indeed done: */
    while ( 1 )
    {
        allDone = 1;
        for ( i = 0; i < numBlocks; i++ )
        {
            if ( (blockHeaders[i]->dwFlags & WHDR_DONE) == 0 )
                allDone = 0;
        }
        if ( allDone )
            break;
        Sleep(20);
    }

    /* Unprepare all mixing blocks: */
    for ( i = 0; i < numBlocks; i++ )
    {
        if ( blockPrepared[i] )
        {
            if ( (mmError = waveOutUnprepareHeader(waveHandle, blockHeaders[i],
                sizeof(WAVEHDR))) != 0 )
                WinError(mmError);
        }
    }

    /* Close wave output device: */
    if ( (mmError = waveOutClose(waveHandle)) != 0 )
        WinError(mmError);

    /* Deallocate all blocks: */
    for ( i = 0; i < numBlocks; i++ )
    {
        free(blocks[i]);
        free(blockHeaders[i]);
    }

    free(blocks);
    free(blockHeaders);
    free(blockPrepared);

    audioInitialized = 0;
}



static void StartAudio(void)
{
    unsigned i;
    unsigned bufferLength;
    unsigned elemSize;
    WAVEFORMATEX format;
    WAVEHDR *header;
    MMRESULT mmError;

    if ( audioInitialized )
        StopAudio();
    
    elemSize = 1;
    if ( stereo )
        elemSize *= 2;
    if ( bits == 16 )
        elemSize *= 2;

    format.wFormatTag = WAVE_FORMAT_PCM;
    if ( stereo )
        format.nChannels = 2;
    else
        format.nChannels = 1;
    format.nSamplesPerSec = sampleRate;
    format.nAvgBytesPerSec = elemSize * sampleRate;
    format.nBlockAlign = (WORD) elemSize;
    if ( bits == 16 )
        format.wBitsPerSample = 16;
    else
        format.wBitsPerSample = 8;
    format.cbSize = 0;

    if ( (mmError = waveOutOpen(&waveHandle, WAVE_MAPPER, &format, 0, 0, 0))
        != 0 )
        WinError(mmError);

    if ( win32AudioBufferLength < 0 )
    {
        bufferLength = (sampleRate * elemSize * BUFFERLEN) / 1000;
    }
    else
    {
        bufferLength = (unsigned) win32AudioBufferLength;
    }

    blockLen = BLOCKSIZE;
    numBlocks = (bufferLength + blockLen - 1) / blockLen;

    blocks = malloc(numBlocks * sizeof(uchar*));
    blockHeaders = malloc(numBlocks * sizeof(WAVEHDR*));
    blockPrepared = malloc(numBlocks * sizeof(int));

    for ( i = 0; i < numBlocks; i++ )
    {
        blocks[i] = (uchar*) malloc(blockLen);
        header = blockHeaders[i] = (WAVEHDR*) malloc(sizeof(WAVEHDR));        

        header->lpData = (char*) blocks[i];
        header->dwBufferLength = blockLen;
        header->dwFlags = WHDR_DONE;        /* mark the block is done */
        header->dwLoops = 0;

        /* Block header is not prepared: */
        blockPrepared[i] = 0;
    }

    audioInitialized = 1;
    blockNum = 0;
    blockPos = 0;
}


void audio_info_struct_init(struct audio_info_struct *ai)
{
  ai->rate = -1;
  ai->gain = -1;
  ai->output = -1;
  ai->device = NULL;
  ai->channels = -1;
  ai->format = -1;
}


int audio_get_formats(struct audio_info_struct *ai)
{
  return AUDIO_FORMAT_SIGNED_16 | AUDIO_FORMAT_UNSIGNED_8;
}


int audio_rate_best_match(struct audio_info_struct *ai)
{
  return 0;
}


int audio_reset_parameters(struct audio_info_struct *ai)
{
  int ret;
  ret = audio_set_format(ai);
  if(ret >= 0)
    ret = audio_set_channels(ai);
  if(ret >= 0)
    ret = audio_set_rate(ai);
  return ret;
}


int audio_set_format(struct audio_info_struct *ai)
{
    switch(ai->format) {
        case -1:
        case AUDIO_FORMAT_SIGNED_16:
        default:
            bits = 16;
            StartAudio();
            break;

        case AUDIO_FORMAT_UNSIGNED_8:
            bits = 8;
            StartAudio();
            break;
    }

    return 0;
}



int audio_open(struct audio_info_struct *ai)
{
    if(!ai)
        return -1;

    stereo = ai->channels - 1;

    if(ai->rate == -1)
        sampleRate = 44100;
    else
        sampleRate = ai->rate;

    bits = 16;

    StartAudio();

    return 0;
}

int audio_set_rate(struct audio_info_struct *ai)
{
    if(ai->rate >= 0)
    {
        sampleRate = ai->rate;
        StartAudio();
    }
    
    return 0;
}


int audio_set_channels(struct audio_info_struct *ai)
{
    stereo = ai->channels - 1;
    StartAudio();
    return 0;
}


int audio_play_samples(struct audio_info_struct *ai,unsigned char *buf,
                       int len)
{
    unsigned doNow;
    unsigned spaceFree;
    MMRESULT mmError;
    unsigned size = len;
    uchar *buffer = (uchar*) buf;
    
    if ( !audioInitialized )
        StartAudio();

    while ( size )
    {
        while ( (blockHeaders[blockNum]->dwFlags & WHDR_DONE) && size )
        {
            if ( blockPrepared[blockNum] )
            {
                if ( (mmError = waveOutUnprepareHeader(waveHandle,
                                                       blockHeaders[blockNum],
                                                       sizeof(WAVEHDR))) != 0 )
                    WinError(mmError);
                blockPrepared[blockNum] = 0;
            }
            
            spaceFree = blockLen - blockPos;
            if ( spaceFree > size )
                doNow = size;
            else
                doNow = spaceFree;

            memcpy(&blocks[blockNum][blockPos], buffer, doNow);
            blockPos += doNow;
            buffer += doNow;
            size -= doNow;

            if ( blockPos >= blockLen )
            {
                blockHeaders[blockNum]->dwFlags = 0;
                blockHeaders[blockNum]->lpData = (char*) blocks[blockNum];
                blockHeaders[blockNum]->dwBufferLength = blockLen;
                blockHeaders[blockNum]->dwFlags = 0;
                blockHeaders[blockNum]->dwLoops = 0;

                if ( (mmError = waveOutPrepareHeader(waveHandle,
                                                     blockHeaders[blockNum],
                                                     sizeof(WAVEHDR))) != 0 )
                    WinError(mmError);
                blockPrepared[blockNum] = 1;

                if ( (mmError = waveOutWrite(waveHandle,
                                             blockHeaders[blockNum],
                                             sizeof(WAVEHDR))) != 0 )
                    WinError(mmError);

                blockPos = 0;
                blockNum++;
                if ( blockNum >= numBlocks )
                    blockNum = 0;
            }            
        }

        if ( size )
        {
            Sleep(10);
        }
    }

    return 2*len;
}


int audio_close(struct audio_info_struct *ai)
{
    StopAudio();
    return 0;
}



void win32AudioFlush(void)
{
    unsigned i;
    int notDone;

    do
    {
        notDone = 0;
        for ( i = 0; i < numBlocks; i++ )
        {
            if ( (blockHeaders[i]->dwFlags & WHDR_DONE) == 0 )
                notDone = 1;
        }
        if ( notDone )
            Sleep(10);
    } while ( notDone );
}
